#include <stdio.h>
#include <stdlib.h>
#include "graph.h"
#include "fifo.h"
#include "limits.h"


static int *mark;
static Stack Q;

#define   gcc        true
#if !defined(DBL_MIN)
#define DBL_MIN      -1.e30
#endif

#if !defined(DBL_MAX)
#define DBL_MAX     1.e30
#endif

#if !defined(INT_MIN)
#define INT_MIN     -32767
#endif

/* Compile the program with gcc if gcc= true and with an ansi compiler
   otherwise */

#if gcc == true
double drand48();
void   srand48(long int);

#define random(A) (int) (drand48()*(double)(A))
#define srandom(A)   srand48(A)
#else
#define srandom(A)   srand(A)

#define random(A)   (int) ((double) rand() / (double) RAND_MAX * (double) (A))
#endif

#define RANDOM(A, B) (int)(random(B-A + 1) + (A))
#define newline  printf("\n")

#define Vector(type, dim)  (type*) calloc((unsigned) dim,(unsigned) sizeof(type))
#define empty       -1


static double *cost_for_Dijsktra;
static double *dist_for_Dijsktra;
static int *prec_for_Dijsktra;
static int *visited;

extern void init_int_fifo(int size, Stack *stack);
extern void empiler(int number, Stack *stack);
extern void depiler(int *number, Stack *stack);




print_graph(Graph * N)
{
  int    i;
  printf("nb of vertices= %i nb of edges= %i\n", N->n, N->m);
  for (i = 0; i < N->m; i++)
    printf("%3i %7i\n", N->or[i], N->ex[i]);
}

void   init_graphe(Graph * N, int nb_of_vertices, int nb_of_edges)
{
  N->n = nb_of_vertices;
  N->m = nb_of_edges;
  N->la = Vector(int, 2 * nb_of_edges);
  N->ls = Vector(int, 2 * nb_of_edges);
  N->lp = Vector(int, nb_of_vertices + 1);
  N->or = Vector(int, nb_of_edges + 1);
  N->ex = Vector(int, nb_of_edges + 1);
}

void   print_demand(Graph * N, double *capacities, int nb_of_demand,
		           Demand * demand, double flow_max)
{
  int    i;
  printf(" %3i  %7i %12i \n", N->m, N->n, nb_of_demand);
  for (i = 0; i < N->m; i++) {
    capacities[i] += flow_max;
    capacities[i] *= 1.3;
    printf("%3i %5i %15lf\n", N->or[i], N->ex[i], capacities[i]);
    /* printf("%lf", cost_for_Dijsktra[i]); */
  }

  for (i = 0; i < nb_of_demand; i++)
    printf("%3i %7i %12i\n", demand[i].or, demand[i].ex, (int) demand[i].val);
}



void   adjacency_list(Graph * N)
{

  int    i, j;
  int    nb = 0;

  N->lp[0] = 0;

  if (N->oriented == true) {
    for (i = 0; i < N->n; i++) {
      for (j = 0; j < N->m; j++) {
	if (N->or[j] == i) {
	  N->la[nb] = j;
	  nb++;
	}
      }
      N->lp[i + 1] = nb;
    }
  }
  else {
    for (i = 0; i < N->n; i++) {
      for (j = 0; j < N->m; j++) {
	if (N->or[j] == i || N->ex[j] == i) {
	  N->la[nb] = j;
	  nb++;
	}
      }
      N->lp[i + 1] = nb;
    }
  }
}

void   succesor_list(Graph * N)
{
  int    i, k, u;

  for (k = 0; k < N->n; k++) {
    for (i = N->lp[k]; i < N->lp[k + 1]; i++) {
      u = N->la[i];
      if ((N->ls[i] = N->or[u]) == k)
	N->ls[i] = N->ex[u];
    }
  }
}

void   init_check_connectivity(int n_vertex)
{
  int    stack_limit;
  stack_limit = n_vertex * 10;
  mark = Vector(int, n_vertex);
  init_int_fifo(stack_limit, &Q);
}




int    check_connectivity(Graph * N)
{
  int    nb_of_vertex = N->n;
  int    i;
  int    neighboor, vertex, connected;


  for (i = 0; i < nb_of_vertex; i++)
    mark[i] = false;
  Free_stack(Q);

  vertex = 0;
  mark[vertex] = true;
  empiler(vertex, &Q);
  while (!empty_stack(Q)) {
    depiler(&vertex, &Q);
    mark[vertex] = true;
    for (i = N->lp[vertex]; i < N->lp[vertex + 1]; i++) {
      neighboor = N->ls[i];
      if (mark[neighboor] == false)
	empiler(neighboor, &Q);
    }
  }
  connected = true;
  for (i = 0; i < nb_of_vertex && connected == true; i++)
    if (!mark[i])
      connected = false;
  return connected;
}


void   GenConnectedGraph(Graph * N)
/* Generate a  connected graph with n_vert vertices and n_edge edges */
{
  int    i, j;
  int    n_vert, n_edge, not_exist, edge1, or1, ex1;
  int    connected;


  n_vert = N->n;
  n_edge = N->m;


  if (n_vert < 2 || n_edge < 0 || n_edge < n_vert ||
      n_edge > max_edge_graph(n_vert)) {
    printf("illegal parameter\n");
    exit(1);
  }
  for (i = 0; i < n_edge; i++) {
    do {
      not_exist = true;
      N->or[i] = random(n_vert);
      do {
	N->ex[i] = random(n_vert);
      } while (N->or[i] == N->ex[i]);


      for (j = 0; j < i; j++)
	if (EdgeEqp(N, i, j))
	  not_exist = false;
    } while (not_exist == false);
  }

  /* we have generated a graph and we check that it is connected */
  adjacency_list(N);
  succesor_list(N);

  connected = check_connectivity(N);

  if (connected == false) {

    /* the graph is not connected. We will permute the edges untill is it
       connected */
    do {
      do {
	edge1 = random(n_edge);
	or1 = random(n_vert);
	do {
	  ex1 = random(n_vert);
	} while (ex1 == or1);
	N->or[edge1] = or1;
	N->ex[edge1] = ex1;
	not_exist = true;
	for (j = 0; j < n_edge; j++)
	  if (EdgeEqp(N, edge1, j) && j != edge1)
	    not_exist = false;
      } while (not_exist == false);
      adjacency_list(N);
      succesor_list(N);
    } while (check_connectivity(N) == false);
  }
}

Graph *init_GenGraph(int nb_max_vertices, int nb_max_edges, int nb_of_block)
{
  Graph *N;
  int    i;
  N = Vector(Graph, nb_of_block);
  if (N == NULL) {
    printf("not enough memory \n");
    exit(1);
  }
  visited = Vector(int, nb_max_vertices);
  for (i = 0; i < nb_of_block; i++)
    init_graphe(&N[i], nb_max_vertices, nb_max_edges);
  return N;
}







void   GenGraph(Graph *N, Graph *Blocks, int *connector, int nb_of_vertices,
		       int nb_of_edges, int nb_of_block)
{
  int    i, j, rest1, rest2;

  if (nb_of_block < 2) {
    printf("illegal parameter\n");
    exit(1);
  }
  Blocks[0].n = nb_of_block - 1;
  Blocks[0].m = Blocks[0].n +
    random(max_edge_graph(Blocks[0].n) - Blocks[0].n);
  rest1 = nb_of_edges - Blocks[0].m - (nb_of_block - 1) *
    (int) ((nb_of_edges - Blocks[0].m) / (nb_of_block - 1));
  rest2 = nb_of_vertices - (nb_of_block - 1) *
    (int) (nb_of_vertices / (nb_of_block - 1));

  GenConnectedGraph(&(Blocks[0]));

  for (i = 1; i < nb_of_block; i++) {
    Blocks[i].n = (int) (nb_of_vertices / (nb_of_block - 1));
    if (i <= rest2)
      Blocks[i].n++;
    Blocks[i].m = (int) ((nb_of_edges - Blocks[0].m) / (nb_of_block - 1));
    if (i <= rest1)
      Blocks[i].m++;
    GenConnectedGraph(&(Blocks[i]));
  }


  /* we build the graph from its blocks Block[1], Block[2], ... */
  N->n = 0;
  N->m = 0;


  for (i = 1; i < nb_of_block; i++) {
    connector[i - 1] = N->n;
    for (j = 0; j < Blocks[i].m; j++) {
      N->ex[j + N->m + Blocks[0].m] = Blocks[i].ex[j] + N->n;
      N->or[j + N->m + Blocks[0].m] = Blocks[i].or[j] + N->n;
    }
    N->n += Blocks[i].n;
    N->m += Blocks[i].m;
  }
  for (j = 0; j < Blocks[0].m; j++) {
    N->ex[j] = connector[Blocks[0].ex[j]];
    N->or[j] = connector[Blocks[0].or[j]];
  }
  N->m += Blocks[0].m;
}







void   init_gen_flow(int nb_of_vertices,int nb_of_edges)
{
  cost_for_Dijsktra = Vector(double, nb_of_edges);
  prec_for_Dijsktra = Vector(int, nb_of_edges);
  dist_for_Dijsktra = Vector(double, nb_of_vertices);
  if (cost_for_Dijsktra == NULL || prec_for_Dijsktra == NULL ||
      dist_for_Dijsktra == NULL) {
    printf("not enough memory\n");
    exit(1);
  }
}



void   ErrChem(int is,int it)
{
  printf("There is no path between %i and %i\n", is, it);
  exit(1);
}



void   add_capac( Graph *N, int   *prec, double *capacities, Demand demand)
{
  int    u, som;

  som = demand.ex;
  if (prec[demand.ex] == INT_MIN) {
    print_graph(N);
    ErrChem(demand.or, demand.ex);
  }
  while (som != demand.or) {
    u = prec[som];
    capacities[u] += demand.val;
    if (som == N->ex[u])
      som = N->or[u];
    else
      som = N->ex[u];
  }
}




void   afficher_liste( Graph *H, double *cout)
{
  int    i, j, u, k;
  printf("\n");
  for (i = 0; i < H->n; i++) {
    for (j = H->lp[i]; j < H->lp[i + 1]; j++) {
      u = H->la[j];
      k = H->ls[j];
      printf("%d=  ", u);
      printf("%i %i cout:%lf", i, k, cout[u]);
      printf("\n");
    }
  }
}


void   Dijsktra(Graph *H, double *d, int *prec, Demand demand,
  double *cost_of_edges)
{
  int    i, j, is, it, u, k;
  double z;

  is = demand.or;
  it = demand.ex;

  for (i = 0; i < H->n; i++) {
    d[i] = DBL_MAX;
    prec[i] = INT_MIN;
    visited[i] = false;
  }

  d[is] = 0;
  for (k = H->lp[is]; k < H->lp[is + 1]; k++) {
    u = H->la[k];
    j = H->ls[k];
    prec[j] = u;
    d[j] = cost_of_edges[u];
  }
  visited[is] = true;
  i = is;	   /* d(is)=0, S={is}, i=is */

  /* debut de Dijkstra */

  do {

    z = DBL_MAX;
    i = INT_MIN;
    for (j = 0; j < H->n; j++) {
      if (visited[j] == false && d[j] <= z) {
	i = j;
	z = d[j];
      }
    }
    if (i < 0)
      break;


    for (k = H->lp[i]; k < H->lp[i + 1]; k++) {
      u = H->la[k];
      j = H->ls[k];
      if (visited[j] == false) {
	z = d[i] + cost_of_edges[u];
	if (z < d[j]) {
	  d[j] = z;
	  prec[j] = u;
	}
      }
    }
    visited[i] = true;
  } while (1);
}








void   gen_flow( Graph *N, Demand *demand,  double *capacities,
  int    nb_of_commodities, double flow_min,double flow_max)
{
        /* demand[0..nb_of_commodities-1] */
	/* capacities[0..N->m-1]   */
  int    i;
  if (flow_min <= 0 || flow_max < flow_min) {
    printf("illegal parameter\n");
    exit(1);
  }
  for (i = 0; i < nb_of_commodities; i++) {
    demand[i].or = random(N->n - 1);
    do {
      demand[i].ex = random(N->n - 1);
    } while (demand[i].ex == demand[i].or);
    demand[i].val = (double) RANDOM(flow_min, flow_max);
  }

  for (i = 0; i < N->m; i++)
    cost_for_Dijsktra[i] = (double) RANDOM(flow_min, flow_max);
  for (i = 0; i < nb_of_commodities; i++) {
    Dijsktra(N, dist_for_Dijsktra, prec_for_Dijsktra, demand[i],
	     cost_for_Dijsktra);

    add_capac(N, prec_for_Dijsktra, capacities, demand[i]);
  }
}



void   main(int argc, char *argv[])
{
  int    i;

  Graph  N;
  Graph *Blocks;
  Demand *demand;
  double *capacities;
  double flow_min, flow_max;
  int   *connector;
  int    nb_of_vertices, nb_of_edges, max_nb_of_edges, nb_blocks, nb_of_demand;
  long int seed;

  if (argc < 7) {
    printf("not enough parameter\n");
    exit(1);

  }
  else {
    nb_of_vertices = atoi(argv[1]);
    nb_of_edges = atoi(argv[2]);
    nb_of_demand = atoi(argv[3]);
    nb_blocks = atoi(argv[4]);
    flow_min = atof(argv[5]);
    flow_max = atof(argv[6]);
    seed = atoi(argv[7]);
  }

  if (nb_of_edges < nb_of_vertices - nb_blocks - max_edge_graph(nb_blocks)) {
    printf("not enough edges\n");
    exit(1);
  }
  max_nb_of_edges = 0;
  for (i = 1; i < nb_blocks; i++)
    max_nb_of_edges += max_edge_graph((int) (nb_of_vertices / (nb_blocks - 1)));
  if (nb_of_edges > max_nb_of_edges) {
    printf("too much edges\n");
    exit(1);
  }

  srandom(seed);

  printf("genflot %i %i %i %i %f %f (%u)\n", nb_of_vertices, nb_of_edges,
	 nb_of_demand, nb_blocks, flow_min, flow_max, seed);

  connector = Vector(int, nb_blocks);
  demand = Vector(Demand, nb_of_demand);

  init_graphe(&N, nb_of_vertices, nb_of_edges);
  init_check_connectivity(N.n);
  Blocks = init_GenGraph(nb_of_vertices, nb_of_edges, nb_blocks);
  GenGraph(&N, Blocks, connector, nb_of_vertices, nb_of_edges, nb_blocks);
  capacities = Vector(double, N.m);
  init_gen_flow(N.n, N.m);
  adjacency_list(&N);
  succesor_list(&N);
  gen_flow(&N, demand, capacities, nb_of_demand, flow_min, flow_max);
  print_demand(&N, capacities, nb_of_demand, demand, flow_max);

}
